from config import *

@njit(cache = True)
def binary_search(entities: np.ndarray, _id: int, end: int) -> tuple[int, int]:
    """
        Searches for the given id within the set of entities.

        Parameters:
            entities: set of IDs to search within

            _id: ID to search for

            end: the end index of the array
        
        Returns:
            (flag. position), where flag indicates success.
                0: success, +/-1: failure
    """

    a_pos = 0
    a_flag = entities[a_pos] - _id
    if a_flag < 0:
        a_flag = -1
    else:
        return (a_flag, a_pos)
    
    b_pos = end
    b_flag = entities[b_pos] - _id
    if b_flag > 0:
        b_flag = 1
    else:
        return (b_flag, b_pos)
    
    while (a_flag * b_flag) < 0 and (b_pos - a_pos) > 1:

        c_pos = (a_pos + b_pos) // 2
        c_flag = min(1, max(-1, entities[c_pos] - _id))
        if c_flag > 0:
            c_flag = 1
        elif c_flag < 0:
            c_flag = -1
        else:
            return (0, c_pos)
        
        if (a_flag * c_flag) < 0:
            b_pos = c_pos
            b_flag = c_flag
        else:
            a_pos = c_pos
            a_flag = c_flag
    
    if b_flag == 0:
        return (0, b_pos)
    return (a_flag, a_pos)

@njit(cache = True)
def realloc(old_array: np.ndarray, new_size: int) -> np.ndarray:
    """
        Reallocates an old array to fit the new size.

        Parameters:
            old_array: the array to be resized

            new_size: the requested size, in elements
        
        Returns:
            A resized array, holding the original elements plus extra space
    """

    new_array = np.zeros(shape = new_size, dtype = old_array.dtype)
    
    for i in range(old_array.size):
        new_array[i] = old_array[i]
    
    return new_array

@njit(cache = True)
def insert(
    entities: np.ndarray, components: np.ndarray, size: int, 
    _id: int, component: np.ndarray, stride: int) -> None:
    """
        Insert a new component.

        Parameters:
            entities: an array of entities

            components: an array of components

            size: the number of elements in the entity array
                currently in use
            
            _id: the id of the entity being inserted

            component: the component being inserted

            stride: the number of elements per component
    """

    i = size
    while i > 0 and entities[i - 1] >= _id:
        entities[i] = entities[i - 1]
        components[stride * i : stride * (i + 1)] = \
            components[stride * (i - 1) : stride * i]
        i -= 1
    entities[i] = _id
    components[stride * i : stride * (i + 1)] = component[:]

@njit(cache = True)
def delete(array: np.ndarray, i: int, size: int, stride: int) -> None:
    """
        Delete a component from an array.

        Parameters:

            array: the array from which to delete

            i: the position of the component to delete

            size: the number of components occupying the array

            stride: the number of ints/floats per component
    """

    while i < (size - 1):
        array[stride * i : stride * (i + 1)] = array[stride * (i + 1) : stride * (i + 2)]
        i += 1

class ComponentSet:
    """
        A Generic container set that can hold data and
        interface with the GPU.
    """

    def __init__(self, data_type: np.dtype, stride: int):
        """
            Initialize a component set.

            Parameters:
                data_type: the type for the numpy array.
                    Only homogenous data types are currently supported.
                
                stride: the number of elements per component.
        """

        self._data_type = data_type
        self._stride = stride
        self._size = 0
        self._capacity = 0

        self._component_array = np.array([], dtype = self._data_type)
        self._entity_array = np.array([], dtype = np.int32)

        self._component_memory = glGenBuffers(1)
        self._entity_memory = glGenBuffers(1)
    
    def insert(self, _id: int, component: np.ndarray) -> None:
        """
            Insert the given component.

            Parameters:

                _id: the id of the component to insert

                component: the data being inserted
        """

        if self._size >= self._capacity:
            self._capacity = 2 * max(1, self._capacity)
            self._component_array = realloc(
                self._component_array, self._stride * self._capacity)
            self._entity_array = realloc(self._entity_array, self._capacity)
            self.rebuild_buffers()

        insert(
            self._entity_array, self._component_array, self._size,
            _id, component, self._stride)
        self._size += 1
        self.upload()

    def delete(self, _id) -> None:
        """
            Delete a component.

            Parameters:

                _id: the id of the component to be deleted.
        """

        flag, i = binary_search(self._entity_array, _id, self._size - 1)
        if (flag != 0):
            return

        delete(self._entity_array, i, self._size, 1)
        delete(self._component_array, i, self._size, self._stride)

        self._size -= 1
    
    def upload(self) -> None:
        """
            Upload the contents of the numpy arrays
            to the GPU's SSBOs
        """
        
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._component_memory)
        glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, 
            self._component_array.nbytes, self._component_array)
        
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._entity_memory)
        glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, 
            self._entity_array.nbytes, self._entity_array)
    
    def rebuild_buffers(self) -> None:
        """
            Delete the current buffers, create new ones,
            and set their storage size.
        """

        #delete old buffers
        glDeleteBuffers(2, (self._component_memory, self._entity_memory))
        #allocate new buffers
        (self._component_memory, self._entity_memory) = glGenBuffers(2)

        #declare storage size etc
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._component_memory)
        glBufferStorage(
            GL_SHADER_STORAGE_BUFFER, 4 * self._stride * self._capacity, 
            None, GL_DYNAMIC_STORAGE_BIT)
        
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._entity_memory)
        glBufferStorage(
            GL_SHADER_STORAGE_BUFFER, 4 * self._capacity, 
            None, GL_DYNAMIC_STORAGE_BIT)
        
    def bind(self, bindings: tuple[int]) -> None:
        """
            Arm the buffers for use.

            Parameters:

                bindings: (component, entity) binding indices.
        """

        glBindBufferBase(
            GL_SHADER_STORAGE_BUFFER, bindings[0], self._component_memory)
        glBindBufferBase(
            GL_SHADER_STORAGE_BUFFER, bindings[1], self._entity_memory)
    
    def destroy(self) -> None:
        """
            Free the memory.
        """

        glDeleteBuffers(2, (self._component_memory, self._entity_memory))

class Buffer:
    """
        Simpler than a ComponentSet, a buffer doesn't keep track of entities.
    """

    def __init__(self, data_type: np.dtype, stride: int):
        """
            Initialize a buffer.

            Parameters:
                data_type: the type for the numpy array.
                    Only homogenous data types are currently supported.
                
                stride: the number of elements per component.
        """

        self._data_type = data_type
        self._stride = stride
        self._capacity = 0

        self._host_memory = np.array([], dtype = self._data_type)

        self._device_memory = glGenBuffers(1)
    
    def insert(self, i: int, component: np.ndarray) -> None:
        """
            Insert the given component.

            Parameters:

                i: the position at which to insert.

                component: the data to insert.
        """

        while i >= self._capacity:
            self._capacity = 2 * max(1, self._capacity)
            self._host_memory = realloc(
                self._host_memory, self._stride * self._capacity)
            self.rebuild()

        self._host_memory[self._stride * i : self._stride * (i + 1)] = component[:]
        self.upload()

    def upload(self) -> None:
        """
            Upload the contents of the numpy array
            to the GPU's SSBO
        """
        
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._device_memory)
        glBufferSubData(GL_SHADER_STORAGE_BUFFER, 0, 
            self._host_memory.nbytes, self._host_memory)
    
    def rebuild(self) -> None:
        """
            Delete the current buffer, create a new one,
            and set its storage size.
        """

        #delete old buffer
        glDeleteBuffers(1, (self._device_memory,))
        #allocate new buffer
        self._device_memory = glGenBuffers(1)

        #declare storage size
        glBindBuffer(GL_SHADER_STORAGE_BUFFER, self._device_memory)
        glBufferStorage(
            GL_SHADER_STORAGE_BUFFER, 4 * self._stride * self._capacity, 
            None, GL_DYNAMIC_STORAGE_BIT)
        
    def bind(self, binding: int) -> None:
        """
            Arm the buffer for use.

            Parameters:

                binding: the index to bind to.
        """

        glBindBufferBase(
            GL_SHADER_STORAGE_BUFFER, binding, self._device_memory)
    
    def destroy(self) -> None:
        """
            Free the memory.
        """

        glDeleteBuffers(1, (self._device_memory,))

class CounterBuffer:
    """
        An atomic counter buffer, zeroed out upon binding and
        modified by the compute shader it's bound to.
    """

    def __init__(self, size: int):
        """
            Initialize a counter buffer.

            Parameters:
                size: the number of things to count.
        """

        self._clear_value = np.zeros(size, dtype = np.uint32)
        self._device_memory = glGenBuffers(1)
        glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, self._device_memory)
        glBufferStorage(
            GL_ATOMIC_COUNTER_BUFFER, self._clear_value.nbytes, 
            None, GL_DYNAMIC_STORAGE_BIT)
        
    def bind(self, binding: int) -> None:
        """
            Arm the buffer for use.

            Parameters:

                binding: the index to bind to.
        """
        
        glBindBuffer(GL_ATOMIC_COUNTER_BUFFER, self._device_memory)
        glBufferSubData(GL_ATOMIC_COUNTER_BUFFER, 0, 
            self._clear_value.nbytes, self._clear_value)
        glBindBufferBase(
            GL_ATOMIC_COUNTER_BUFFER, binding, self._device_memory)
    
    def destroy(self) -> None:
        """
            Free the memory.
        """

        glDeleteBuffers(1, (self._device_memory,))