############################## Imports   ######################################
#region
from config import *
import components
import systems
import factory
#endregion
##################################### Control #################################
#region
class App:
    """
        The control class.
    """


    def __init__(self):
        """
            Initialize the program.
        """

        self._set_up_glfw()
        self._set_up_timer()
        self._create_assets()

    def _set_up_glfw(self) -> None:
        """
            Initialize and configure GLFW
        """

        glfw.init()
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MAJOR,4)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_CONTEXT_VERSION_MINOR,6)
        glfw.window_hint(
            GLFW_CONSTANTS.GLFW_OPENGL_PROFILE, 
            GLFW_CONSTANTS.GLFW_OPENGL_CORE_PROFILE)
        glfw.window_hint(GLFW_CONSTANTS.GLFW_OPENGL_FORWARD_COMPAT, GLFW_CONSTANTS.GLFW_TRUE)
        self._window = glfw.create_window(
            SCREEN_WIDTH, SCREEN_HEIGHT, "Title", None, None)
        glfw.make_context_current(self._window)
        #for uncapped framerate
        glfw.swap_interval(0)
    
    def _set_up_timer(self) -> None:
        """
            Initialize the variables used by the framerate
            timer.
        """

        self._last_time = glfw.get_time()
        self._current_time = 0
        self._frames_rendered = 0
        self._frametime = 0.0
    
    def _create_assets(self) -> None:
        """
            Create all of the assets needed by the program.
        """

        self._make_components()

        self._factory = factory.Factory(
            self._batches, self._transforms, self._velocities, 
            self._model_matrices, self._object_types, 
            self._colors, self._lights)
        
        self._factory.register_batch(OBJECT_TYPE_CUBE)
        self._factory.register_batch(OBJECT_TYPE_LIGHT)

        self._make_cubes()
        self._make_lights()
        self._batches.upload()
        self._make_player()
        
        self._make_systems()
    
    def _make_components(self) -> None:
        """
            Make all the component sets used by the program.
        """

        self._shaders = np.array(
            (factory.create_shader(
                "shaders/vertex.txt", "shaders/fragment.txt"),
            factory.create_shader(
                "shaders/simple_3d_vertex.txt", 
                "shaders/simple_3d_fragment.txt")), dtype=np.uint32)

        #base_instance, instance_count
        self._batches = components.Buffer(data_type=np.uint32, stride=2)
        #x, y, z, _, rx, ry, rz, _
        self._transforms = components.ComponentSet(
            data_type=np.float32, stride=TRANSFORM_STRIDE)
        #x, y, z, _, rx, ry, rz, _
        self._velocities = components.ComponentSet(
            data_type=np.float32, stride=VELOCITY_STRIDE)
        self._model_matrices = components.Buffer(
            data_type=np.float32, stride=MAT4_STRIDE)
        self._object_types = components.ComponentSet(data_type=np.uint32, stride = 1)
        #rgbs
        self._colors = components.ComponentSet(
            data_type=np.float32, stride=COLOR_STRIDE)
        self._lights = components.Buffer(
            data_type=np.float32, stride=LIGHT_STRIDE)

        #pos, eulers, right, up, forwards
        self._camera = np.zeros(PLAYER_STRIDE, dtype = np.float32)
        self._view_matrix = np.zeros(MAT4_STRIDE, dtype = np.float32)
    
    def _make_cubes(self) -> None:
        """
            Make a whole bunch of cubes.
        """

        for _ in range(200):

            x = np.random.uniform(low = -10, high = 10)
            y = np.random.uniform(low = -10, high = 10)
            z = np.random.uniform(low = -10, high = 10)

            e_x = np.random.uniform(low = 0, high = 360)
            e_y = np.random.uniform(low = 0, high = 360)
            e_z = np.random.uniform(low = 0, high = 360)

            ev_x = np.random.uniform(low = -0.2, high = 0.2)
            ev_y = np.random.uniform(low = -0.2, high = 0.2)
            ev_z = np.random.uniform(low = -0.2, high = 0.2)
            
            self._factory.make_cube(
                position = (x, y, z), 
                eulers = (e_x, e_y, e_z), 
                angular_velocity = (ev_x, ev_y, ev_z))
    
    def _make_lights(self) -> None:
        """
            Make a bunch of lights.
        """

        for _ in range(8):

            x = np.random.uniform(low = -10, high = 10)
            y = np.random.uniform(low = -10, high = 10)
            z = np.random.uniform(low = -10, high = 10)

            r = np.random.uniform(low = 0.5, high = 1.0)
            g = np.random.uniform(low = 0.5, high = 1.0)
            b = np.random.uniform(low = 0.5, high = 1.0)
            s = np.random.uniform(low = 2, high = 5)
            
            self._factory.make_light(
                position = (x, y, z), 
                color = (r, g, b, s))
    
    def _make_player(self) -> None:
        """
            The player is tightly coupled to the input,
            and there's only one of them, so they're 
            represented with a few local variables.
        """

        x = -10
        f_x = 1.0
        r_y = -1.0
        u_z = 1.0

        self._camera[0] = x
        self._camera[6] = f_x
        self._camera[10] = r_y
        self._camera[14] = u_z
        
        #load identity into view transform
        for i in range(4):
            self._view_matrix[5 * i] = 1.0
    
    def _make_systems(self) -> None:
        """
            Make all the systems that the engine
            will use.
        """

        self._velocity_system = systems.VelocitySystem(
            self._transforms, self._velocities)

        self._transform_system = systems.TransformSystem(
            self._transforms, self._object_types, 
            self._model_matrices, self._batches)
        
        self._light_system = systems.LightSystem(
            self._transforms, self._colors, self._lights)
        self._light_system.update()

        self._input_system = systems.InputSystem(
            self._window, self._camera, 
            self._view_matrix, self._shaders)
        
        self._render_system = systems.RenderSystem(
            self._shaders, self._batches, 
            self._model_matrices, self._lights)
        
    def run(self) -> None:
        """
            Run the program.
        """

        while (self._input_system.update(self._frametime / 16.67)):

            self._velocity_system.update(self._frametime / 16.67)
            self._transform_system.update()
            
            self._render_system.update()
            glfw.swap_buffers(self._window)

            #timing
            self._calculate_framerate()

    def _calculate_framerate(self) -> None:
        """
            Calculate the framerate and frametime,
            and update the window title.
        """

        self._current_time = glfw.get_time()
        delta = self._current_time - self._last_time
        if (delta >= 1):
            framerate = max(1,int(self._frames_rendered/delta))
            glfw.set_window_title(self._window, f"Running at {framerate} fps.")
            self._last_time = self._current_time
            self._frames_rendered = -1
            self._frametime = float(1000.0 / max(1,framerate))
        self._frames_rendered += 1
    
    def quit(self):
        """
            Destroy everything!
        """
        
        glDeleteProgram(self._shaders[0])
        glDeleteProgram(self._shaders[1])
        self._render_system.destroy()
#endregion
###############################################################################