############################## Imports   ######################################
#region
from config import *
import factory
import components
#endregion
############################## Constants ######################################
#region
GLOBAL_X = np.array([1,0,0], dtype=np.float32)
GLOBAL_Y = np.array([0,1,0], dtype=np.float32)
GLOBAL_Z = np.array([0,0,1], dtype=np.float32)
COMPUTE_THRESHOLD = 512
#endregion
############################## Transform System ###############################
#region
@njit(cache = True)
def update_transforms(
    transform_components: np.ndarray, 
    transform_entities: np.ndarray, 
    object_type_components: tuple[np.ndarray],
    object_type_entities: tuple[np.ndarray],
    model_matrices: np.ndarray,
    batches: np.ndarray,
    counts: tuple[int]) -> None:
    """
        Write the transform data to matrices.

        Parameters:
            transforms: describes the objects (their position and rotation)

            object_types: integer code designating the object type

            model_matrices: stores the objects' model transform matrices

            batches: describe the rendering batches (base_instance, instance_count)

            counts: the number of: (transforms, objects)
    """

    cubes_written = 0
    lights_written = 0
    for i in range(counts[0]):

        entity = transform_entities[i]

        write_index = 0
        _, j = components.binary_search(
            object_type_entities, entity, counts[1] - 1)
        object_type = object_type_components[j]
        if object_type == 0:
            write_index = cubes_written
            cubes_written += 1
        else:
            base_instance = batches[2]
            write_index = base_instance + lights_written
            lights_written += 1

        #unpack transform data
        index = TRANSFORM_STRIDE * i
        x     = transform_components[index + TRANSFORM_OFFSET_X]
        y     = transform_components[index + TRANSFORM_OFFSET_Y]
        z     = transform_components[index + TRANSFORM_OFFSET_Z]
        ey    = transform_components[index + TRANSFORM_OFFSET_EY]
        ez    = transform_components[index + TRANSFORM_OFFSET_EZ]

        r_y = np.radians(ey)
        r_z = np.radians(ez)
        c_y = np.cos(r_y)
        s_y = np.sin(r_y)
        c_z = np.cos(r_z)
        s_z = np.sin(r_z)

        #write transforms
        index = MAT4_STRIDE * write_index

        model_matrices[index + MAT4_R1C1] = c_y * c_z
        model_matrices[index + MAT4_R2C1] = c_y * s_z
        model_matrices[index + MAT4_R3C1] = -s_y

        model_matrices[index + MAT4_R1C2] = -s_z
        model_matrices[index + MAT4_R2C2] = c_z

        model_matrices[index + MAT4_R1C3] = s_y * c_z
        model_matrices[index + MAT4_R2C3] = s_y * s_z
        model_matrices[index + MAT4_R3C3] = c_y

        model_matrices[index + MAT4_R1C4] = x
        model_matrices[index + MAT4_R2C4] = y
        model_matrices[index + MAT4_R3C4] = z
        model_matrices[index + MAT4_R4C4] = 1.0

class TransformSystem:
    """
        Consumes Transform components and builds model matrices.
    """

    
    def __init__(self, transforms: components.ComponentSet,  
        object_types: components.ComponentSet,
        model_matrices: components.Buffer,
        batches: components.Buffer):
        """
            Initialize a new transform system.

            Parameters:

                transforms: transform components (position and rotation)

                object_types: object type components (integer codes)

                model_matrices: to be populated with model transforms

                batches: describe the rendering batches for each object type
        """
        
        self._transforms = transforms
        self._object_types = object_types
        self._models = model_matrices
        self._batches = batches
        
        self._transform_bindings = (0, 1)
        self._object_bindings = (2, 3)
        self._batch_binding = 4
        self._model_binding = 5
        self._counter_binding = 6
        
        self._shader = factory.create_compute_shader("shaders/update_transforms.txt")
        
        self._counts = np.zeros(2, dtype=np.int32)
        glUseProgram(self._shader)
        self._counts_location = glGetUniformLocation(self._shader, "counts")
        
        self._counter_buffer = components.CounterBuffer(2)
    
    def update(self) -> None:
        """
            Build the model matrices from the transform info.
        """

        self._counts[0] = self._transforms._size
        self._counts[1] = self._object_types._size
        
        if self._counts[0] >= COMPUTE_THRESHOLD:
            glUseProgram(self._shader)
            glUniform2iv(self._counts_location, 1, self._counts)
            self._transforms.bind(self._transform_bindings)
            self._object_types.bind(self._object_bindings)
            self._batches.bind(self._batch_binding)
            self._models.bind(self._model_binding)
            self._counter_buffer.bind(self._counter_binding)
            workgroup_count = int(np.ceil(self._counts[0] / 64))
            glDispatchCompute(workgroup_count, 1, 1)
            glMemoryBarrier(
                GL_SHADER_STORAGE_BARRIER_BIT | GL_ATOMIC_COUNTER_BARRIER_BIT)
        else:
            update_transforms(
                self._transforms._component_array, 
                self._transforms._entity_array, 
                self._object_types._component_array, 
                self._object_types._entity_array, 
                self._models._host_memory, 
                self._batches._host_memory, self._counts)
            self._models.upload()
        
#endregion
############################## Velocity System ################################
#region
@njit(cache = True)
def update_velocities(
    velocity_components: tuple[np.ndarray],
    velocity_entities: tuple[np.ndarray],
    transform_components: np.ndarray, 
    transform_entities: np.ndarray, 
    counts: tuple[int], dt: float) -> None:
    """
        Updates the position and rotation of objects based on their
        linear and angular velocities.

    Parameters:
        
        velocity: linear and angular velocities of objects
        
        transform: positions and rotations of objects
        
        counts: number of (velocity, transform) components
        
        dt: framerate correction factor
    """

    for i in range(counts[0]):

        base_index = VELOCITY_STRIDE * i
        v_x = dt * velocity_components[base_index + VELOCITY_OFFSET_X]
        v_y = dt * velocity_components[base_index + VELOCITY_OFFSET_Y]
        v_z = dt * velocity_components[base_index + VELOCITY_OFFSET_Z]
        e_x = dt * velocity_components[base_index + VELOCITY_OFFSET_EX]
        e_y = dt * velocity_components[base_index + VELOCITY_OFFSET_EY]
        e_z = dt * velocity_components[base_index + VELOCITY_OFFSET_EZ]

        entity = velocity_entities[i]

        _, j = components.binary_search(
            transform_entities, entity, counts[1] - 1)
        
        base_index = TRANSFORM_STRIDE * j

        transform_components[base_index + TRANSFORM_OFFSET_X] = \
            transform_components[base_index + TRANSFORM_OFFSET_X] + v_x
        transform_components[base_index + TRANSFORM_OFFSET_Y] = \
            transform_components[base_index + TRANSFORM_OFFSET_Y] + v_y
        transform_components[base_index + TRANSFORM_OFFSET_Z] = \
            transform_components[base_index + TRANSFORM_OFFSET_Z] + v_z

        transform_components[base_index + TRANSFORM_OFFSET_EX] = \
            (transform_components[base_index + TRANSFORM_OFFSET_EX] + e_x) % 360
        transform_components[base_index + TRANSFORM_OFFSET_EY] = \
            (transform_components[base_index + TRANSFORM_OFFSET_EY] + e_y) % 360
        transform_components[base_index + TRANSFORM_OFFSET_EZ] = \
            (transform_components[base_index + TRANSFORM_OFFSET_EZ] + e_z) % 360

class VelocitySystem:
    """
        Uses velocity components to update transform components
    """

    
    def __init__(self, transforms: components.ComponentSet,
        velocities: components.ComponentSet):
        """
            Initialize a new velocity system.

        Parameters:
            transforms: Set of transform components
            velocities: Set of velocity components
        """
        
        self._transforms = transforms
        self._velocities = velocities
        
        self._velocity_bindings = (0, 1)
        self._transform_bindings = (2, 3)
        
        self._shader = factory.create_compute_shader("shaders/update_velocities.txt")
        
        self._counts = np.zeros(2, dtype=np.int32)
        glUseProgram(self._shader)
        self._counts_location = glGetUniformLocation(self._shader, "counts")
        self._dt_location = glGetUniformLocation(self._shader, "dt")
        
    def update(self, dt: float) -> None:
        """
            Use the velocity components to update transform components.
        
        Parameters:
        
            dt: framerate correction factor
        """

        self._counts[0] = self._velocities._size
        self._counts[1] = self._transforms._size
        if self._counts[0] >= COMPUTE_THRESHOLD:
            glUseProgram(self._shader)
            glUniform2iv(self._counts_location, 1, self._counts)
            glUniform1f(self._dt_location, dt)
            self._velocities.bind(self._velocity_bindings)
            self._transforms.bind(self._transform_bindings)
            workgroup_count = int(np.ceil(self._counts[0] / 64))
            glDispatchCompute(workgroup_count, 1, 1)
            glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT)

        else:
            update_velocities(
                self._velocities._component_array, 
                self._velocities._entity_array, 
                self._transforms._component_array, 
                self._transforms._entity_array, self._counts, dt)
        
#endregion
############################## Light System ###################################
#region
@njit(cache = True)
def update_lights(
    color_components: tuple[np.ndarray],
    color_entities: tuple[np.ndarray],
    transform_components: np.ndarray, 
    transform_entities: np.ndarray, 
    lights: np.ndarray,
    counts: tuple[int]) -> None:
    """
        Update the light components.

    Args:
        
        color: rgb color and strength
        
        transform: position and rotation
        
        lights: light component buffer to write to
        
        counts: number of (color, transform) components
    """

    for i in range(counts[0]):

        index = COLOR_STRIDE * i
        r = color_components[index + COLOR_OFFSET_R]
        g = color_components[index + COLOR_OFFSET_G]
        b = color_components[index + COLOR_OFFSET_B]
        s = color_components[index + COLOR_OFFSET_S]

        entity = color_entities[i]

        _, j = components.binary_search(
            transform_entities, entity, counts[1] - 1)
        
        index = TRANSFORM_STRIDE * j

        x = transform_components[index + TRANSFORM_OFFSET_X]
        y = transform_components[index + TRANSFORM_OFFSET_Y]
        z = transform_components[index + TRANSFORM_OFFSET_Z]

        base_index = LIGHT_STRIDE * i
        lights[base_index + LIGHT_OFFSET_X] = x
        lights[base_index + LIGHT_OFFSET_Y] = y
        lights[base_index + LIGHT_OFFSET_Z] = z
        lights[base_index + LIGHT_OFFSET_R] = r
        lights[base_index + LIGHT_OFFSET_G] = g
        lights[base_index + LIGHT_OFFSET_B] = b
        lights[base_index + LIGHT_OFFSET_S] = s

class LightSystem:
    """
        Writes color and position info to light buffers.
    """

    
    def __init__(self, transforms: components.ComponentSet,
        colors: components.ComponentSet, 
        lights: components.Buffer):
        """
            Initializes a light system

        Parameters:
            transforms: Set of transform components.
            
            colors: Set of color components.
            
            lights: Light buffer.
        """
        
        self._transforms = transforms
        self._colors = colors
        self._lights = lights
    
    def update(self) -> None:
        """
            Update the light buffer.
        """

        counts = (self._colors._size, self._transforms._size)

        update_lights(
            self._colors._component_array, 
            self._colors._entity_array, 
            self._transforms._component_array, 
            self._transforms._entity_array,
            self._lights._host_memory, counts)
        self._lights.upload()
#endregion
############################## Input System ###################################
#region
@njit(cache = True)
def move_player(player: np.ndarray, d_pos: np.ndarray) -> None:
    """
        Move the player.
        
        Parameters:
        
            player: the player's position and orientation data
            
            d_pos: coefficients, (d_forwards, d_right, d_up)
    """

    x = player[PLAYER_OFFSET_X]
    y = player[PLAYER_OFFSET_Y]
    z = player[PLAYER_OFFSET_Z]

    x += d_pos[0] * player[PLAYER_OFFSET_FX]
    y += d_pos[0] * player[PLAYER_OFFSET_FY]
    z += d_pos[0] * player[PLAYER_OFFSET_FZ]

    x += d_pos[1] * player[PLAYER_OFFSET_RX]
    y += d_pos[1] * player[PLAYER_OFFSET_RY]
    z += d_pos[1] * player[PLAYER_OFFSET_RZ]

    x += d_pos[2] * player[PLAYER_OFFSET_UX]
    y += d_pos[2] * player[PLAYER_OFFSET_UY]
    z += d_pos[2] * player[PLAYER_OFFSET_UZ]

    player[PLAYER_OFFSET_X] = x
    player[PLAYER_OFFSET_Y] = y
    player[PLAYER_OFFSET_Z] = z

@njit(cache = True)
def spin_player(player: np.ndarray, d_eulers: np.ndarray) -> None:
    """
        Spin the player.
        
        Parameters:
        
            player: Data describing the player's orientation.
            
            d_eulers: amount to move about the (x, y, z) axes.
    """

    e_x = player[PLAYER_OFFSET_EX] + d_eulers[0]
    e_y = player[PLAYER_OFFSET_EY] + d_eulers[1]
    e_z = player[PLAYER_OFFSET_EZ] + d_eulers[2]


    player[PLAYER_OFFSET_EX] = e_x % 360
    player[PLAYER_OFFSET_EY] = min(89, max(-89, e_y))
    player[PLAYER_OFFSET_EZ] = e_z % 360

@njit(cache = True)
def update_view(camera: np.ndarray, view: np.ndarray) -> None:
    """
        Update the camera's orientation and view transform.
        
        Parameters:
        
            camera: Camera's orientation data.
            
            view: Camera's view matrix.
    """

    e_y = camera[PLAYER_OFFSET_EY]
    e_z = camera[PLAYER_OFFSET_EZ]

    c_y = np.cos(np.radians(e_y))
    s_y = np.sin(np.radians(e_y))
    c_z = np.cos(np.radians(e_z))
    s_z = np.sin(np.radians(e_z))

    f_x = c_z * c_y
    f_y = s_z * c_y
    f_z = s_y
    norm = np.sqrt(f_x * f_x + f_y * f_y + f_z * f_z)
    f_x = f_x / norm
    f_y = f_y / norm
    f_z = f_z / norm

    r_x = f_y
    r_y = -f_x
    r_z = 0.0
    norm = np.sqrt(r_x * r_x + r_y * r_y + r_z * r_z)
    r_x = r_x / norm
    r_y = r_y / norm
    r_z = r_z / norm

    u_x = r_y * f_z - r_z * f_y
    u_y = r_z * f_x - r_x * f_z
    u_z = r_x * f_y - r_y * f_x
    norm = np.sqrt(u_x * u_x + u_y * u_y + u_z * u_z)
    u_x = u_x / norm
    u_y = u_y / norm
    u_z = u_z / norm

    camera[PLAYER_OFFSET_FX] = f_x
    camera[PLAYER_OFFSET_FY] = f_y
    camera[PLAYER_OFFSET_FZ] = f_z
    camera[PLAYER_OFFSET_RX] = r_x
    camera[PLAYER_OFFSET_RY] = r_y
    camera[PLAYER_OFFSET_RZ] = r_z
    camera[PLAYER_OFFSET_UX] = u_x
    camera[PLAYER_OFFSET_UY] = u_y
    camera[PLAYER_OFFSET_UZ] = u_z

    x = camera[PLAYER_OFFSET_X]
    y = camera[PLAYER_OFFSET_Y]
    z = camera[PLAYER_OFFSET_Z]

    view[MAT4_R1C1] = r_x
    view[MAT4_R2C1] = u_x
    view[MAT4_R3C1] = -f_x

    view[MAT4_R1C2] = r_y
    view[MAT4_R2C2] = u_y
    view[MAT4_R3C2] = -f_y
    
    view[MAT4_R1C3] = r_z
    view[MAT4_R2C3] = u_z
    view[MAT4_R3C3] = -f_z

    view[MAT4_R1C4] = -(r_x * x + r_y * y + r_z * z)
    view[MAT4_R2C4] = -(u_x * x + u_y * y + u_z * z)
    view[MAT4_R3C4] = f_x * x + f_y * y + f_z * z
    view[MAT4_R4C4] = 1.0

class InputSystem:
    """
        Handles mouse and keyboard events/state.
    """

    
    def __init__(self, 
        window: glfw._GLFWwindow, 
        camera: np.ndarray, 
        view_matrix: np.ndarray, 
        shaders: np.ndarray):
        """
            Initialize a new input system

        Parameters:
            window: The window to listen to.
            
            camera: Camera's orientation data
            
            view_matrix: Camera's view matrix
            
            shaders: Shaders (for uploading camera info)
        """

        self._window = window
        self._camera = camera
        self._view = view_matrix
        self._shaders = shaders

        self._set_up_input_systems()
        self._get_uniform_locations()

    def _set_up_input_systems(self) -> None:
        """
            Configure the mouse and keyboard
        """

        glfw.set_input_mode(
            self._window, 
            GLFW_CONSTANTS.GLFW_CURSOR, 
            GLFW_CONSTANTS.GLFW_CURSOR_HIDDEN
        )

        self._keys = {}
        glfw.set_key_callback(self._window, self._key_callback)
    
    def _key_callback(self, 
        window: glfw._GLFWwindow, key: int, 
        scancode: int, action: int, mods: int) -> None:
        """
            Handle a key event.

            Parameters:

                window: the window on which the keypress occurred.

                key: the key which was pressed

                scancode: scancode of the key

                action: action of the key event

                mods: modifiers applied to the event
        """

        state = False
        match action:
            case GLFW_CONSTANTS.GLFW_PRESS:
                state = True
            case GLFW_CONSTANTS.GLFW_RELEASE:
                state = False
            case _:
                return

        self._keys[key] = state
    
    def _get_uniform_locations(self) -> None:
        """
            Query and store the locations of various uniforms
            on the shaders. During update, the input system is responsible
            for uploading camera info to these locations.
        """

        self._view_location = {}

        pipeline = PIPELINE_TYPE_STANDARD
        shader = self._shaders[pipeline]
        glUseProgram(shader)
        self._view_location[pipeline] = glGetUniformLocation(shader, "view")
        self._camera_pos_location = glGetUniformLocation(shader, "cameraPos")

        pipeline = PIPELINE_TYPE_LIT
        shader = self._shaders[pipeline]
        glUseProgram(shader)
        self._view_location[pipeline] = glGetUniformLocation(shader, "view")

    def update(self, dt: float) -> bool:
        """
            Handle any input events.

        Parameters:
            dt: Framerate correction factor

        Returns:
            bool: Whether the program should end.
        """

        if glfw.window_should_close(self._window) \
            or self._keys.get(GLFW_CONSTANTS.GLFW_KEY_ESCAPE, False):
            
            return False

        self._handle_keys(dt)
        self._handle_mouse()
        glfw.poll_events()

        update_view(self._camera, self._view)

        pipeline = PIPELINE_TYPE_STANDARD
        shader = self._shaders[pipeline]
        glUseProgram(shader)
        glUniformMatrix4fv(self._view_location[pipeline], 1, GL_FALSE, self._view)
        glUniform3fv(self._camera_pos_location, 1, self._camera)

        pipeline = PIPELINE_TYPE_LIT
        shader = self._shaders[pipeline]
        glUseProgram(shader)
        glUniformMatrix4fv(self._view_location[pipeline], 1, GL_FALSE, self._view)

        return True
    
    def _handle_keys(self, dt) -> None:
        """
            Takes action based on the keys currently pressed.
            
            Parameters:
            
                dt: framerate correction factor.
        """

        rate = 0.05 * dt
        d_pos = np.zeros(3, dtype=np.float32)

        if self._keys.get(GLFW_CONSTANTS.GLFW_KEY_W, False):
            d_pos += GLOBAL_X
        if self._keys.get(GLFW_CONSTANTS.GLFW_KEY_A, False):
            d_pos -= GLOBAL_Y
        if self._keys.get(GLFW_CONSTANTS.GLFW_KEY_S, False):
            d_pos -= GLOBAL_X
        if self._keys.get(GLFW_CONSTANTS.GLFW_KEY_D, False):
            d_pos += GLOBAL_Y

        length: float = pyrr.vector.length(d_pos)

        if abs(length) < 0.00001:
            return

        d_pos = rate / length * d_pos
        move_player(self._camera, d_pos)

    def _handle_mouse(self) -> None:
        """
            spin the player based on the mouse movement
        """

        (x,y) = glfw.get_cursor_pos(self._window)
        d_eulers = 0.02 * ((SCREEN_WIDTH / 2) - x) * GLOBAL_Z
        d_eulers += 0.02 * ((SCREEN_HEIGHT / 2) - y) * GLOBAL_Y
        spin_player(self._camera, d_eulers)
        glfw.set_cursor_pos(self._window, SCREEN_WIDTH / 2, SCREEN_HEIGHT / 2)
#endregion
############################## Render System ##################################
#region
class RenderSystem:
    """
        Draws everything.
    """
    

    def __init__(self, 
        shaders: np.ndarray, 
        batches: components.Buffer, 
        models: components.Buffer,
        lights: components.Buffer):
        """
            Initializes the rendering system.
            
            Parameters:
            
                shaders: shaders used for rendering
                
                batches: instance batch descriptions
                
                models: model matrix storage buffer
                
                lights: light storage buffer
        """

        self._shaders = shaders
        self._batches = batches
        self._models = models
        self._lights = lights

        self._set_up_opengl()

        self._create_assets()

        self._set_onetime_uniforms()

    def _set_up_opengl(self) -> None:
        """
            Configure any desired OpenGL options
        """

        glClearColor(0.1, 0.1, 0.1, 1)
        glEnable(GL_DEPTH_TEST)
        glEnable(GL_CULL_FACE)
        glCullFace(GL_BACK)
    
    def _create_assets(self) -> None:
        """
            Create all of the assets needed for drawing.
        """

        self._vbos = {}
        self._vaos = {}
        self._vertex_counts = {}
        self._materials = {}
        
        object_type = OBJECT_TYPE_CUBE
        vbo, vao, vertex_count = factory.load_mesh("models/cube.obj")
        self._vbos[object_type] = vbo
        self._vaos[object_type] = vao
        self._vertex_counts[object_type] = vertex_count
        
        albedo = factory.load_material("gfx/crate_albedo.png")
        ao = factory.load_material("gfx/crate_ao.png")
        normal = factory.load_material("gfx/crate_normal.png")
        specular = factory.load_material("gfx/crate_specular.png")
        self._materials[object_type] = (albedo, ao, normal, specular)
        
        object_type = OBJECT_TYPE_LIGHT
        vbo, vao, vertex_count = factory.make_cuboid_mesh(
            l = 0.1, w = 0.1, h = 0.1)
        self._vbos[object_type] = vbo
        self._vaos[object_type] = vao
        self._vertex_counts[object_type] = vertex_count

    def _set_onetime_uniforms(self) -> None:
        """
            Some shader data only needs to be set once.
        """

        projection_transform = pyrr.matrix44.create_perspective_projection(
            fovy = 45, aspect = 640/480, 
            near = 0.1, far = 50, dtype=np.float32
        )

        pipeline = PIPELINE_TYPE_STANDARD
        shader = self._shaders[pipeline]
        glUseProgram(shader)

        glUniformMatrix4fv(
            glGetUniformLocation(shader,"projection"),
            1, GL_FALSE, projection_transform
        )

        glUniform3fv(
            glGetUniformLocation(shader,"ambient"), 
            1, np.array([0.1, 0.1, 0.1],dtype=np.float32))

        glUniform1i(
            glGetUniformLocation(shader, "material.albedo"), 0)

        glUniform1i(
            glGetUniformLocation(shader, "material.ao"), 1)

        glUniform1i(
            glGetUniformLocation(shader, "material.normal"), 2)

        glUniform1i(
            glGetUniformLocation(shader, "material.specular"), 3)
        
        pipeline = PIPELINE_TYPE_LIT
        shader = self._shaders[pipeline]
        glUseProgram(shader)

        glUniformMatrix4fv(
            glGetUniformLocation(shader,"projection"),
            1, GL_FALSE, projection_transform
        )
      
    def update(self) -> None:
        """
            Draw everything!
        """

        #refresh screen
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        #bind storage buffers
        self._models.bind(0)
        self._lights.bind(1)

        pipeline = PIPELINE_TYPE_STANDARD
        shader = self._shaders[pipeline]
        glUseProgram(shader)

        object_type = OBJECT_TYPE_CUBE
        materials = self._materials[object_type]
        for i in range(4):
            glActiveTexture(GL_TEXTURE0 + i)
            glBindTexture(GL_TEXTURE_2D, materials[i])
        glBindVertexArray(self._vaos[object_type])
        base_instance = self._batches._host_memory[2 * object_type]
        instance_count = self._batches._host_memory[2 * object_type + 1]
        glDrawArraysInstancedBaseInstance(
            GL_TRIANGLES, 0, 
            self._vertex_counts[object_type], instance_count, base_instance)

        pipeline = PIPELINE_TYPE_LIT
        shader = self._shaders[pipeline]
        glUseProgram(shader)

        object_type = OBJECT_TYPE_LIGHT
        glBindVertexArray(self._vaos[object_type])
        base_instance = self._batches._host_memory[2 * object_type]
        instance_count = self._batches._host_memory[2 * object_type + 1]
        glDrawArraysInstancedBaseInstance(
            GL_TRIANGLES, 0, 
            self._vertex_counts[object_type], instance_count, base_instance)

    def destroy(self) -> None:
        """ free any allocated memory """

        glDeleteVertexArrays(2, 
            (
                self._vaos[OBJECT_TYPE_CUBE], 
                self._vaos[OBJECT_TYPE_LIGHT]))
        glDeleteBuffers(2, 
            (
                self._vbos[OBJECT_TYPE_CUBE], 
                self._vbos[OBJECT_TYPE_LIGHT]))
        glDeleteTextures(6, self._materials[OBJECT_TYPE_CUBE])
#endregion
###############################################################################